#!/usr/bin/perl

###############################################################################
#
# This file can be distributed and modified freely, as long as this original
# header is not modified, and the signature header
#      "X-ExtScanner: Niversoft's FindAttachments (free)
# still added to scanned messages
#
# Author:         Nicolas Hatier <nicolas dot hatier at niversoft dot com>
# Organization:   Niversoft ides logicielles / software ideas
# Date:           June 2004
#
# This code is provided AS IS, without any certification that it will even run
#
###############################################################################
#
# This script is as a Content-Filtering script for Communigate Pro.
# The purpose of this script is to find attachements in mails transmitted within
# the server and add headers with the extensions of these attachements to help
# identifying them with a server-wide or an account rule.
#
# This script uses a custom, fast algorithm to search attached files. It's not
# supposed to miss any, but may report attached files that's not really there
# in very specific case, such as a message containing MIME headers samples in
# its body. However, this is rarely bound to happend.
#
# This script doesn't have any dependency except Perl itself.
#
# Add this filter in Helpers. This file is saved with DOS line endings, some
# shells may have problems reading the #! line at the beginning of the script
# It's recommended to specfy the interpreter in the Helper program path:
#     /usr/bin/perl Filters/Find_Attachments.pl
#
# Recent versions of perl doesn't have problem with line endings.
#
# Then the helper is set, add rule to make all incoming messages pass through
# this filter. This rule should have no condition and the action:
#    ExternalFilter  Find_Attachments
#
# Then add a rule like this if you want to ban some extensions:
#    if HEADER is "X-AttachExt: exe"
#    reject with "This server doesn't accept exe attached files"
#
# This last rule action and condition can be modified to follow your needs.
#
# Note that cgpro rules are not case-sensitive
#
# v1.23: Now decodes quoted-printable and base64
# v1.24: A few fixes to avoid looping or missing attachments on truncated or
#        badly-formatted MIME messages (thanks to Scott Burch)
# v1.25: Fixed a bug introduced in 1.24 that could lead the script to miss
#        some attachments

use strict;

$| = 1;
my %attached;

main();

sub Stringify
{
   my $headers = join "\\e", @_;
   $headers =~ s|\\|\\\\|g;
   $headers =~ s|"|\\"|g;
   $headers =~ s|\\\\e|\\e|g;

   return $headers;
}

sub getBase64Code
{
   my $code = ord(substr($_[0], $_[1], 1));
   return $code - 65 if (($code >= 65) && ($code <= 90));
   return $code - 71 if (($code >= 97) && ($code <= 122));
   return $code +  4 if (($code >= 48) && ($code <= 57));
   return 62 if ($code == 43);
   return 63 if ($code == 47);
   return 0;
}

sub Base64
{
	my $txt = "";
	my $len = length $_[0];
	for (my $i = 0; $i < $len; $i += 4)
	{
		my @chs = (getBase64Code($_[0], $i + 0), getBase64Code($_[0], $i + 1), getBase64Code($_[0], $i + 2), getBase64Code($_[0], $i + 3));
	   my $tot0 = (($chs[0] << 2) + ($chs[1] >> 4)) % 256;
	   my $tot1 = (($chs[1] << 4) + ($chs[2] >> 2)) % 256;
	   my $tot2 = (($chs[2] << 6) + ($chs[3] >> 0)) % 256;
	   $txt .= ($tot0 > 0 ? chr($tot0) : "") . ($tot1 > 0 ? chr($tot1) : "") . ($tot2 > 0 ? chr($tot2) : "");
	}
  return $txt;
}

sub QuotedPrintable
{
   $_ = $_[0];
   s/=([\da-fA-F]{2})/pack("C", hex($1))/ge;
   return $_;
}


sub FindAttachements
{
   #print "* $_[0]";

   if ($_[0] =~ /\W$_[1] *= *"(.*)"/)
   {
      #print "*! $1\n";
      $attached{$1} = 1;
   }
   elsif ($_[0] =~ /\W$_[1] *= *([^;\r\n]*)/) # includes badly formatted MIME header
   {
      #print "*! $1\n";
      $attached{$1} = 1;
   }
}

sub main
{
   print "* FindAttachments Filter (free), v1.25\n";
   while (<>)
   {
      my @File;
      my ($ID, $Command, @File) = split / |\r|\n/;
      my $File = join " ", @File;

      print "$ID OK\n"      and next if ($Command =~ /^quit$/i);
	   print "$ID INTF 3\n"  and next if ($Command =~ /^intf$/i);
	   print "$ID OK\n"      and next if ($Command =~ /^key$/i);
		print "$ID FAILURE\n" and next if ($Command !~ /^file$/i); # other unsupported commands

      if (open(FILE, $File))
      {
         %attached = ();
         my $line = <FILE>; chomp($line);
         while($line ne '')
         {
          $line = <FILE>;
          chomp($line);
         }

         $line = <FILE>;
         while (defined $line)
         {
            if ($line =~ /^content-type:/i)
            {
               FindAttachements($line, "name");
               my $subline = <FILE>;
               while (defined $subline)
               {
                  if ($subline =~ /^[ \t]/)
                  {
                     FindAttachements($subline, "name");
                  }
                  else
                  {
                     $line = $subline;
                     last;
                  }
                  $subline = <FILE>;
               }
            }
            if ($line =~ /^content-disposition:/i)
            {
               FindAttachements($line, "filename");
               my $subline = <FILE>;
               while (defined $subline)
               {
                  if ($subline =~ /^[ \t]/)
                  {
                     FindAttachements($subline, "filename");
                  }
                  else
                  {
                     $line = $subline;
                     last;
                  }
                  $subline = <FILE>;
               }
            }
            $line = <FILE>;
         }
         close FILE;

         my %exts;
         my @headers;
         my $count = scalar keys %attached;

         if ($count > 0)
         {
            foreach my $filename (keys %attached)
            {
               $filename = QuotedPrintable($1) if ($filename =~ /\?q\?(.*?)\?=/i);
               $filename = Base64($1) if ($filename =~ /\?b\?(.*?)\?=/i);

               if ($filename =~ /\.([^\.]*)$/)
               {
                  $exts{$1} = 1;
               }
               else
               {
                  $exts{"..."} = 1;
               }
            }
            print "* Found $count attached files\n";
         }
         else
         {
            print "* Found no attached files\n";
         }

         foreach (keys %exts)
         {
            push @headers, "X-AttachExt: ".lc($_);
         }
         push @headers, "X-ExtScanner: Niversoft's FindAttachments (free)";
         print "$ID ADDHEADER \"".Stringify(@headers)."\"\n";
      }
      else
      {
            print "* Error file not found\n";
            print "$ID OK\n";
      }
      open STDOUT, ">&STDOUT";
   }
}


